# Copyright 2015 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.

from coal_mine.business_logic import BusinessLogic
from coal_mine.memory_store import MemoryStore
from coal_mine.server import application, url_prefix
import json
from unittest import TestCase
from urllib.parse import urlencode


class ServerTests(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.store = MemoryStore()
        cls.logic = BusinessLogic(cls.store, 'example@example.com')

    @staticmethod
    def environ(path, params):
        return {
            'PATH_INFO': path,
            'QUERY_STRING': urlencode(params),
        }

    @staticmethod
    def make_url(command):
        return url_prefix + command

    def start_response(self, status_code, headers):
        self.response_code = status_code
        self.response_headers = headers

    def call_application(self, url, params, auth_key=None):
        self.response_code = self.response_headers = None
        iterator = application(self.logic, auth_key, self.environ(url, params),
                               self.start_response)
        data = ''.join(d.decode('utf-8') for d in iterator)
        return json.loads(data) if data else None

    def test_create(self):
        self.call_application(
            self.make_url('create'),
            {'name': 'test_create', 'periodicity': 12345})
        self.assertEqual(self.response_code, '200 OK')

    def test_create_name_missing(self):
        self.call_application(
            self.make_url('create'), {'periodicity': 12345})
        self.assertEqual(self.response_code, '400 Bad Request')

    def test_trigger(self):
        created = self.call_application(
            self.make_url('create'),
            {'name': 'test_trigger', 'periodicity': 12346})
        self.call_application('/{}'.format(created['canary']['id']), {})
        self.assertEqual(self.response_code, '200 OK')
        self.call_application(
            '/{}'.format(created['canary']['id']),
            {'comment': 'test_trigger trigger comment'})
        self.assertEqual(self.response_code, '200 OK')

    def test_url_invalid(self):
        self.call_application('/froodle', {})
        self.assertEqual(self.response_code, '404 Not Found')
        self.call_application(self.make_url('froodle'), {})
        self.assertEqual(self.response_code, '404 Not Found')

    def test_auth(self):
        auth_key = 'key_for_test_auth'
        self.call_application(
            self.make_url('create'),
            {'name': 'test_auth', 'periodicity': 12347, 'auth_key': auth_key},
            auth_key=auth_key)
        self.assertEqual(self.response_code, '200 OK')

    def test_auth_missing(self):
        self.call_application(self.make_url('create'), {}, auth_key='froodle')
        self.assertEqual(self.response_code, '401 Unauthorized')

    def test_auth_invalid(self):
        self.call_application(
            self.make_url('create'),
            {'name': 'test_auth_invalid', 'periodicity': 12348,
             'auth_key': 'bad_key'},
            auth_key='good_key')
        self.assertEqual(self.response_code, '401 Unauthorized')

    def test_find_identifier_slug(self):
        self.call_application(
            self.make_url('create'),
            {'name': 'test_find_identifier_slug', 'periodicity': 12349})
        self.call_application(
            self.make_url('get'),
            {'slug': 'test-find-identifier-slug'})
        self.assertEqual(self.response_code, '200 OK')

    def test_find_identifier_name(self):
        self.call_application(
            self.make_url('create'),
            {'name': 'test_find_identifier_name', 'periodicity': 12350})
        self.call_application(
            self.make_url('get'),
            {'name': 'test_find_identifier_name'})
        self.assertEqual(self.response_code, '200 OK')

    def test_find_identifier_missing(self):
        response = self.call_application(self.make_url('get'), {})
        self.assertRegexpMatches(response['error'],
                                 'Must specify id, slug, or name')
        self.assertEqual(self.response_code, '400 Bad Request')
        response = self.call_application(self.make_url('update'), {})
        self.assertRegexpMatches(response['error'], 'Must specify id or slug')
        self.assertEqual(self.response_code, '400 Bad Request')

    def test_boolean_parameters(self):
        self.call_application(
            self.make_url('create'),
            {'name': 'test_boolean_parameters', 'periodicity': 12351,
             'paused': 'true'})
        self.assertEqual(self.response_code, '200 OK')
        self.call_application(
            self.make_url('create'),
            {'name': 'test_boolean_parameters2', 'periodicity': 12352,
             'paused': 'false'})
        self.assertEqual(self.response_code, '200 OK')
        self.call_application(
            self.make_url('create'),
            {'name': 'test_boolean_parameters3', 'periodicity': 12352,
             'paused': 'froodle'})
        self.assertEqual(self.response_code, '400 Bad Request')

    def test_invalid_parameter(self):
        self.call_application(
            self.make_url('create'),
            {'name': 'test_handle_delete', 'periodicity': 12352,
             'froodle': 'freedle'})
        self.assertEqual(self.response_code, '400 Bad Request')

    def test_handle_delete(self):
        self.call_application(
            self.make_url('create'),
            {'name': 'test_handle_delete', 'periodicity': 12353})
        self.assertEqual(self.response_code, '200 OK')
        self.call_application(
            self.make_url('delete'), {'name': 'test_handle_delete'})
        self.assertEqual(self.response_code, '200 OK')

    def test_handle_update(self):
        response = self.call_application(
            self.make_url('create'),
            {'name': 'test_handle_update', 'periodicity': 12354,
             'email': 'test_handle_update@example.com'})
        self.assertEqual(self.response_code, '200 OK')
        self.call_application(
            self.make_url('update'),
            {'id': response['canary']['id'], 'periodicity': 12355,
             'email': ''})
        self.assertEqual(self.response_code, '200 OK')
        self.call_application(
            self.make_url('update'),
            {'id': response['canary']['id'], 'email': '-'})
        self.assertEqual(self.response_code, '200 OK')

    def test_handle_list(self):
        self.call_application(self.make_url('list'), {})
        self.assertEqual(self.response_code, '200 OK')

    def test_handle_pause(self):
        response = self.call_application(
            self.make_url('create'),
            {'name': 'test_handle_pause', 'periodicity': 12356})
        self.assertEqual(self.response_code, '200 OK')
        self.call_application(
            self.make_url('pause'),
            {'id': response['canary']['id']})
        self.assertEqual(self.response_code, '200 OK')

    def test_handle_unpause(self):
        response = self.call_application(
            self.make_url('create'),
            {'name': 'test_handle_unpause', 'periodicity': 12357})
        self.assertEqual(self.response_code, '200 OK')
        self.call_application(
            self.make_url('pause'),
            {'id': response['canary']['id']})
        self.assertEqual(self.response_code, '200 OK')
        self.call_application(
            self.make_url('unpause'),
            {'id': response['canary']['id']})
        self.assertEqual(self.response_code, '200 OK')

    def test_scheduled_periodicity(self):
        self.call_application(
            self.make_url('create'),
            {'name': 'test_scheduled_periodicity',
             'periodicity': '* * * * sat,sun 600; * * * * mon-fri 90'})
        self.assertEqual(self.response_code, '200 OK')

    def test_not_found(self):
        self.call_application(
            self.make_url('trigger'),
            {'name': 'test_not_found'})
        self.assertEqual(self.response_code, '404 Not Found')



# Copyright 2015 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.

from coal_mine.business_logic import (
    AlreadyExistsError,
    AlreadyPausedError,
    AlreadyUnpausedError,
    CanaryNotFoundError,
    BusinessLogic,
)
from coal_mine.memory_store import MemoryStore
from coal_mine.mongo_store import MongoStore
from datetime import datetime, timedelta
import signal
import smtplib
import time
from unittest import TestCase
from unittest.mock import patch
import uuid


class MemoryStoreTester(object):
    def get_store(self):
        return MemoryStore()

    def free_store(self):
        pass


class MongoStoreTester(object):
    def get_store(self):
        self.db_hosts = ['localhost']
        self.db_name = "coal-mine-test-" + str(uuid.uuid4())
        return MongoStore(self.db_hosts, self.db_name, None, None)

    def free_store(self):
        self.store.db.client.drop_database(self.db_name)


class BusinessLogicTests(object):
    def setUp(self):
        self.store = self.get_store()
        self.logic = BusinessLogic(self.store, 'example@example.com')

    def tearDown(self):
        signal.alarm(0)
        signal.signal(signal.SIGALRM, signal.SIG_DFL)
        self.free_store()

    def test_noop(self):
        # Just tests that setUp() and tearDown() don't crash.
        pass

    def test_create(self):
        created = self.logic.create(name='test_create', periodicity=12345)
        fetched = self.logic.get(created['id'])
        self.assertEqual(created, fetched)

    def test_create_invalid(self):
        with self.assertRaises(TypeError):
            self.logic.create(name=2, periodicity=12346)
        with self.assertRaises(TypeError):
            self.logic.create(name='', periodicity=12346)
        self.logic.create(name='test_create_invalid', periodicity=12346)
        with self.assertRaises(AlreadyExistsError):
            self.logic.create(name='test_create_invalid', periodicity=12346)
        with self.assertRaises(TypeError):
            self.logic.create(name='test_create_invalid2', periodicity='abc')
        with self.assertRaises(TypeError):
            self.logic.create(name='test_create_invalid2', periodicity=-1)
        with self.assertRaises(TypeError):
            self.logic.create(name='test_create_invalid2', periodicity=12346,
                              description=2)
        with self.assertRaises(TypeError):
            self.logic.create(name='test_create_invalid2', periodicity=12346,
                              emails='test_create_invalid@example.com')
        with self.assertRaises(TypeError):
            self.logic.create(name='test_create_invalid2', periodicity=12346,
                              paused='abc')

    def test_create_emails_none(self):
        self.logic.create(name='test_create_emails_none',
                          periodicity=12347,
                          emails=None)

    def test_update(self):
        created = self.logic.create(name='test_update',
                                    periodicity=12347)
        self.logic.update(created['id'],
                          name='test_update2',
                          periodicity=12348,
                          description='test_update2 description',
                          emails=['test_update@example.com'])
        fetched = self.logic.get(created['id'])
        self.assertEqual(fetched['name'], 'test_update2')
        self.assertEqual(fetched['periodicity'], 12348)
        self.assertEqual(fetched['description'], 'test_update2 description')
        self.assertEqual(fetched['emails'], ['test_update@example.com'])

    def test_update_same_slug(self):
        created = self.logic.create(name='test_update_same_slug',
                                    periodicity=5)
        self.logic.update(created['id'], name='Test_Update_Same_Slug')
        fetched = self.logic.get(created['id'])
        self.assertEqual(fetched['name'], 'Test_Update_Same_Slug')
        self.assertEqual(created['slug'], fetched['slug'])

    def test_update_paused(self):
        created = self.logic.create('test_update_paused', periodicity=5)
        self.logic.pause(created['id'])
        self.logic.update(created['id'], periodicity=10)
        fetched = self.logic.get(created['id'])
        self.assertNotIn('deadline', fetched)

    def test_update_invalid(self):
        created = self.logic.create(name='test_update_invalid',
                                    periodicity=12349)
        with self.assertRaises(TypeError):
            self.logic.update(created['id'], name=2)
        with self.assertRaises(TypeError):
            self.logic.update(created['id'], name='')
        self.logic.create(name='test_update_invalid2', periodicity=12350)
        with self.assertRaises(AlreadyExistsError):
            self.logic.update(created['id'], name='test_update_invalid2')
        with self.assertRaises(TypeError):
            self.logic.update(created['id'], periodicity='abc')
        with self.assertRaises(TypeError):
            self.logic.update(created['id'], periodicity=-1)
        with self.assertRaises(TypeError):
            self.logic.update(created['id'], description=2)
        with self.assertRaises(TypeError):
            self.logic.update(created['id'],
                              emails='test_update_invalid@example.com')
        with self.assertRaises(ValueError):
            self.logic.update(created['id'])

    def test_update_late_change(self):
        created = self.logic.create(name='test_update_late_change',
                                    periodicity=12351)
        time.sleep(1)
        self.logic.update(created['id'], periodicity=1)
        fetched = self.logic.get(created['id'])
        # Note that this test is mostly for code coverage, but we should at
        # least check that the change we expected is there.
        self.assertNotEqual(created['periodicity'], fetched['periodicity'])

    def test_update_not_found(self):
        with self.assertRaises(CanaryNotFoundError):
            self.logic.update('testunfo', name='test_update_not_found')

    def test_store_unset(self):
        created = self.logic.create('foo', 20)
        self.logic.pause(created['id'])

    def test_trigger(self):
        created = self.logic.create(name='test_trigger', periodicity=12352)
        self.logic.trigger(created['id'])
        self.logic.trigger(created['id'], comment='test_trigger comment')

    def test_trigger_late(self):
        created = self.logic.create(name='test_trigger_late', periodicity=1)
        time.sleep(1.1)
        self.logic.trigger(created['id'])

    def test_trigger_paused(self):
        created = self.logic.create(name='test_trigger_paused',
                                    periodicity=12353,
                                    paused=True)
        self.logic.trigger(created['id'])

    def test_trigger_not_found(self):
        with self.assertRaises(CanaryNotFoundError):
            self.logic.trigger('testtnfo')

    def test_pause(self):
        created = self.logic.create(name='test_pause', periodicity=1)
        time.sleep(1.1)
        self.logic.pause(created['id'])
        with self.assertRaises(AlreadyPausedError):
            self.logic.pause(created['id'])
        self.logic.unpause(created['id'])
        with self.assertRaises(AlreadyUnpausedError):
            self.logic.unpause(created['id'])
        self.logic.pause(created['id'], comment='test_pause pause comment')
        self.logic.unpause(created['id'],
                           comment='test_pause unpause comment')

    def test_pause_not_found(self):
        with self.assertRaises(CanaryNotFoundError):
            self.logic.pause('testpnfo')

    def test_unpause_not_found(self):
        with self.assertRaises(CanaryNotFoundError):
            self.logic.unpause('testunfo')

    def test_delete(self):
        created = self.logic.create(name='test_delete', periodicity=12354)
        self.logic.get(created['id'])
        self.logic.delete(created['id'])
        with self.assertRaises(CanaryNotFoundError):
            self.logic.get(created['id'])

    def test_delete_not_found(self):
        with self.assertRaises(CanaryNotFoundError):
            self.logic.delete('testdnfo')

    def test_list(self):
        self.logic.list()

    def test_list_no_paused_canaries(self):
        self.logic.create('not-paused', 20)
        self.assertEqual(next(self.logic.list())['name'], 'not-paused')
        self.assertEqual(next(self.logic.list(paused=False))['name'],
                         'not-paused')
        with self.assertRaises(StopIteration):
            next(self.logic.list(paused=True))

    def test_list_only_paused_canary(self):
        self.logic.create('paused', 20, paused=True)
        self.assertEqual(next(self.logic.list())['name'], 'paused')
        self.assertEqual(next(self.logic.list(paused=True))['name'],
                         'paused')
        with self.assertRaises(StopIteration):
            next(self.logic.list(paused=False))

    def test_list_paused_and_unpaused_canary(self):
        self.logic.create('not-paused', 10)
        self.logic.create('paused', 20, paused=True)
        iterator = self.logic.list()
        self.assertEqual(set((next(iterator)['name'], next(iterator)['name'])),
                         set(('not-paused', 'paused')))
        iterator = self.logic.list(paused=True)
        self.assertEqual(next(iterator)['name'], 'paused')
        with self.assertRaises(StopIteration):
            next(iterator)
        iterator = self.logic.list(paused=False)
        self.assertEqual(next(iterator)['name'], 'not-paused')
        with self.assertRaises(StopIteration):
            next(iterator)

    def test_list_no_late_canaries(self):
        self.logic.create('not-late', 20)
        self.assertEqual(next(self.logic.list())['name'], 'not-late')
        self.assertEqual(next(self.logic.list(late=False))['name'],
                         'not-late')
        with self.assertRaises(StopIteration):
            next(self.logic.list(late=True))

    def test_list_only_late_canary(self):
        self.logic.create('late', 1)
        time.sleep(1.1)
        self.assertEqual(next(self.logic.list())['name'], 'late')
        self.assertEqual(next(self.logic.list(late=True))['name'],
                         'late')
        with self.assertRaises(StopIteration):
            next(self.logic.list(late=False))

    def test_list_late_and_not_late_canary(self):
        self.logic.create('late', 1)
        self.logic.create('not-late', 20)
        time.sleep(1.1)
        iterator = self.logic.list()
        self.assertEqual(set((next(iterator)['name'], next(iterator)['name'])),
                         set(('not-late', 'late')))
        iterator = self.logic.list(late=True)
        self.assertEqual(next(iterator)['name'], 'late')
        with self.assertRaises(StopIteration):
            next(iterator)
        iterator = self.logic.list(late=False)
        self.assertEqual(next(iterator)['name'], 'not-late')
        with self.assertRaises(StopIteration):
            next(iterator)

    def test_list_search(self):
        self.logic.create('foo', 20)
        next(self.logic.list(search='foo'))
        with self.assertRaises(StopIteration):
            next(self.logic.list(search='froodlefreedle'))
        next(self.logic.list(verbose=True))

    def test_notify(self):
        with patch('smtplib.SMTP'):
            created = self.logic.create(name='test_notify',
                                        periodicity=1,
                                        emails=['test_notify@example.com'])
            time.sleep(1.1)
            self.logic.trigger(created['id'])
        with patch.object(smtplib.SMTP, 'connect', side_effect=Exception):
            time.sleep(1.1)
            self.logic.trigger(created['id'])
        # Not needed for test, but let's clean up after ourselves to avoid
        # unwanted notifications while other tests are running!
        self.logic.delete(created['id'])

    def test_find_identifier(self):
        created = self.logic.create(name='test_find_identifier',
                                    periodicity=12355)
        self.assertEqual(created['id'],
                         self.logic.find_identifier(identifier=created['id']))
        self.assertEqual(created['id'],
                         self.logic.find_identifier(name=created['name']))

    def test_find_identifier_invalid(self):
        with self.assertRaisesRegexp(Exception, 'Must specify'):
            self.logic.find_identifier()
        with self.assertRaisesRegexp(Exception, 'Specify only one'):
            self.logic.find_identifier(name='foo', slug='bar')

    def test_find_identifier_slug_not_found(self):
        with self.assertRaisesRegexp(
                CanaryNotFoundError,
                r"'slug': 'test-find-identifier-slug-not-found'"):
            self.logic.find_identifier(
                slug='test-find-identifier-slug-not-found')

    def test_add_history(self):
        history = []
        self.logic.add_history(history, None)
        for i in range(1000):
            self.logic.add_history(history, str(i))

    def test_add_history_invalid(self):
        history = []
        with self.assertRaises(TypeError):
            self.logic.add_history(history, 2)

    def test_schedule_next_deadline(self):
        # Make sure StopIteration is handled properly when there are no
        # active canaries.
        self.logic.schedule_next_deadline()

    def test_periodicity_numeric(self):
        created = self.logic.create(name='test_periodicity_numeric',
                                    periodicity=1200)
        delta = (created['deadline'] - datetime.utcnow()).total_seconds()
        self.assertAlmostEqual(delta / 10, 120, places=0)

    def test_periodicity_schedule_inactive(self):
        now = datetime.utcnow()
        midnight_tomorrow = (now + timedelta(days=1)).replace(
            hour=0, minute=0, second=0, microsecond=0)
        tomorrow_schedule = '* * * * {} 1200'.format(
            midnight_tomorrow.isoweekday())
        created = self.logic.create(name='test_periodicity_schedule_inactive',
                                    periodicity=tomorrow_schedule)
        delta = (created['deadline'] - midnight_tomorrow).total_seconds()
        self.assertAlmostEqual(delta / 10, 120, places=0)

    def test_periodicity_schedule_active(self):
        now = datetime.utcnow()
        created = self.logic.create(name='test_periodicity_schedule_active',
                                    periodicity='* * * * * 1200')
        delta = (created['deadline'] - now).total_seconds()
        self.assertAlmostEqual(delta / 10, 120, places=0)

    def test_periodicity_invalid(self):
        with self.assertRaises(TypeError):
            self.logic.create(name='test_periodicity_invalid',
                              periodicity='* * * * 1200')

    def test_periodicity_invalid_newline(self):
        with self.assertRaises(TypeError):
            self.logic.create(name='test_periodicity_invalid_newline',
                              periodicity='* * * * sat 1200\n* * * * sun 400')

    def test_periodicity_invalid_command(self):
        with self.assertRaises(TypeError):
            self.logic.create(name='test_periodicity_invalid_command',
                              periodicity='* * * * * froodle')

    def test_periodicity_invalid_negative(self):
        with self.assertRaises(TypeError):
            self.logic.create(name='test_periodicity_invalid_negative',
                              periodicity='* * * * * -1')

    def test_periodicity_invalid_overlapping(self):
        with self.assertRaises(TypeError):
            self.logic.create(name='test_periodicity_invalid_overlapping',
                              periodicity='* * * * * 30; * * * * * 60')

    def test_periodicity_delta_case_2(self):
        periodicity = '* 0 * * * 120'
        whence = datetime(2016, 6, 30, 1, 0)
        delta = self.logic.calculate_periodicity_delta(periodicity, whence)
        next_deadline = whence + delta
        self.assertEqual(next_deadline, datetime(2016, 7, 1, 0, 2))

    def test_periodicity_delta_case_3(self):
        periodicity = '* 0 * * * 120'
        whence = datetime(2016, 6, 30, 0, 59)
        delta = self.logic.calculate_periodicity_delta(periodicity, whence)
        next_deadline = whence + delta
        self.assertEqual(next_deadline, datetime(2016, 7, 1, 0, 2))

    def test_periodicity_delta_case_4(self):
        periodicity = '* 0 * * * 120; * 1 * * * 600'
        whence = datetime(2016, 6, 30, 0, 59)
        delta = self.logic.calculate_periodicity_delta(periodicity, whence)
        next_deadline = whence + delta
        self.assertEqual(next_deadline, datetime(2016, 6, 30, 1, 9))

    def test_deadline_handler_next_deadline(self):
        self.logic.create(name='sooner', periodicity=1)
        later = self.logic.create(name='later', periodicity=2)
        time.sleep(1.1)
        next_deadline = next(self.store.upcoming_deadlines())
        self.assertEqual(later['name'], next_deadline['name'])


class BusinessLogicMemoryTests(MemoryStoreTester, BusinessLogicTests,
                               TestCase):
    pass


class BusinessLogicMongoTests(MongoStoreTester, BusinessLogicTests,
                              TestCase):
    pass



#!/usr/bin/env python

# Copyright 2015 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.

"""
Coal Mine CLI
"""

import argparse
from configparser import SafeConfigParser
import copy
import os
import pprint
import re
import requests
try:
    from simplejson.errors import JSONDecodeError as JSONError
except ImportError:
    JSONError = ValueError
import sys

config_file = '~/.coal-mine.ini'
config_section = 'coal-mine'


def parse_args(args, config_file):
    config = SafeConfigParser()
    config.read([config_file])
    try:
        section = config[config_section]
    except KeyError:
        config['coal-mine'] = {}
        section = config['coal-mine']

    connect_parser = argparse.ArgumentParser(add_help=False)
    host_default = section.get('host', 'localhost')
    connect_parser.add_argument('--host', action='store',
                                help="Server host name or URL (default {})".
                                format(host_default), default=host_default)
    port_default = section.get('port', None)
    connect_parser.add_argument('--port', action='store', type=int,
                                help='Server port', default=port_default)
    auth_key_group = connect_parser.add_mutually_exclusive_group()
    auth_key_default = section.get('auth-key', None)
    auth_key_group.add_argument('--auth-key', action='store',
                                help='Authentication key (default {})'.format(
                                    '<hidden>' if auth_key_default else None),
                                default=auth_key_default)
    auth_key_group.add_argument('--no-auth-key', action='store_true',
                                help='Disable authentication',
                                default=False)

    parser = argparse.ArgumentParser(description="CLI wrapper for Coal "
                                     "Mine's HTTP API")
    subparsers = parser.add_subparsers()

    configure_parser = subparsers.add_parser('configure', help='Save '
                                             'configuration from command line '
                                             'to ' + config_file,
                                             parents=[connect_parser])
    configure_parser.set_defaults(func=handle_configure,
                                  config_parser=config,
                                  config_file=config_file)

    create_parser = subparsers.add_parser('create', help='Create canary',
                                          parents=[connect_parser])
    create_parser.add_argument('--name', action='store', required=True)
    create_parser.add_argument('--periodicity', action='store',
                               type=periodicity, required=True)
    create_parser.add_argument('--description', action='store')
    create_parser.add_argument('--email', action='append')
    create_parser.add_argument('--paused', action='store_true', default=False)
    create_parser.set_defaults(func=handle_create)

    id_parser = argparse.ArgumentParser(add_help=False)
    id_parser_group = id_parser.add_mutually_exclusive_group(required=True)
    id_parser_group.add_argument('--name', action='store')
    id_parser_group.add_argument('--slug', action='store')
    id_parser_group.add_argument('--id', action='store')

    delete_parser = subparsers.add_parser('delete', help='Delete canary',
                                          parents=[connect_parser, id_parser])
    delete_parser.set_defaults(func=handle_delete)

    update_parser = subparsers.add_parser('update', help='Update canary',
                                          parents=[connect_parser])
    update_parser.add_argument('--name', action='store')
    update_parser_group = update_parser.add_mutually_exclusive_group()
    update_parser_group.add_argument('--slug', action='store')
    update_parser_group.add_argument('--id', action='store')
    update_parser.add_argument('--no-history', '--terse', action='store_true',
                               help='Omit history in output')
    update_parser.add_argument('--periodicity', action='store',
                               type=periodicity)
    update_parser.add_argument('--description', action='store')
    update_parser.add_argument('--email', action='append', help='Specify "-" '
                               'to clear existing email(s)')
    update_parser.set_defaults(func=handle_update)

    get_parser = subparsers.add_parser('get', help='Get canary',
                                       parents=[connect_parser, id_parser])
    get_parser.add_argument('--no-history', '--terse', action='store_true',
                            help='Omit history in output')
    get_parser.set_defaults(func=handle_get)

    list_parser = subparsers.add_parser('list', help='List canaries',
                                        parents=[connect_parser])
    list_parser.add_argument('--verbose', action='store_true', default=None)
    list_parser.add_argument('--no-history', '--terse', action='store_true',
                             help='Omit history in output')
    paused_group = list_parser.add_mutually_exclusive_group()
    paused_group.add_argument('--paused', action='store_true', default=None)
    paused_group.add_argument('--no-paused', dest='paused',
                              action='store_false', default=None)
    late_group = list_parser.add_mutually_exclusive_group()
    late_group.add_argument('--late', action='store_true', default=None)
    late_group.add_argument('--no-late', dest='late',
                            action='store_false', default=None)
    list_parser.add_argument('--search', action='store', default=None,
                             help='Regular expression to match against name, '
                             'slug, identifier, and email addresses')
    list_parser.set_defaults(func=handle_list)

    trigger_parser = subparsers.add_parser('trigger', help='Trigger canary',
                                           parents=[connect_parser, id_parser])
    trigger_parser.add_argument('--comment', action='store')
    trigger_parser.set_defaults(func=handle_trigger)

    pause_parser = subparsers.add_parser('pause', help='Pause canary',
                                         parents=[connect_parser, id_parser])
    pause_parser.add_argument('--no-history', '--terse', action='store_true',
                              help='Omit history in output')
    pause_parser.add_argument('--comment', action='store')
    pause_parser.set_defaults(func=handle_pause)

    unpause_parser = subparsers.add_parser('unpause', help='Unpause canary',
                                           parents=[connect_parser, id_parser])
    unpause_parser.add_argument('--no-history', '--terse', action='store_true',
                                help='Omit history in output')
    unpause_parser.add_argument('--comment', action='store')
    unpause_parser.set_defaults(func=handle_unpause)

    args = parser.parse_args(args)

    if 'func' not in args:
        parser.error("No command specified")

    url = ''
    if not re.match(r'^https?:', args.host):
        url += 'http://'
    url += args.host
    if args.port:
        url += ':{}'.format(args.port)
    url += '/coal-mine/v1/canary/'
    args.url = url

    if args.no_auth_key:
        args.auth_key = None
    if args.func is not handle_configure:
        del args.no_auth_key

    return args


def doit(args, config_file):
    args = parse_args(args, config_file)
    args.func(args)


def handle_configure(args):
    section = args.config_parser[config_section]
    section['host'] = args.host
    if args.port:
        section['port'] = str(args.port)
    if args.auth_key:
        section['auth-key'] = args.auth_key
    elif args.no_auth_key:
        section.pop('auth-key', None)
    with open(args.config_file, 'w') as configfile:
        args.config_parser.write(configfile)


def handle_create(args):
    call('create', args)


def handle_delete(args):
    call('delete', args)


def get_no_history_filter(d):
    if 'canary' in d:
        d = copy.deepcopy(d)
        del d['canary']['history']
        return d
    if 'canaries' in d:  # pragma: no cover (should always be true)
        d = copy.deepcopy(d)
        for canary in d['canaries']:
            del canary['history']
        return d
    return d  # pragma: no cover


def handle_update(args):
    if not (args.name or args.id or args.slug):
        sys.exit('Must specify --name, --id, or --slug')
    if args.name and not (args.id or args.slug):
        found = call('get', args, {'name': args.name}, action='return')
        del args.name
        args.id = found['canary']['id']
    if vars(args).pop('no_history', None):
        filter = get_no_history_filter
    else:
        filter = None
    call('update', args, filter=filter)


def handle_get(args):
    if vars(args).pop('no_history', None):
        filter = get_no_history_filter
    else:
        filter = None
    call('get', args, filter=filter)


def handle_list(args):
    if args.paused is None:
        del args.paused
    if args.late is None:
        del args.late
    if args.search is None:
        del args.search
    if vars(args).pop('no_history', None):
        filter = get_no_history_filter
    else:
        filter = None
    call('list', args, filter=filter)


def handle_trigger(args):
    del args.auth_key
    call('trigger', args)


def handle_pause(args):
    if vars(args).pop('no_history', None):
        filter = get_no_history_filter
    else:
        filter = None
    call('pause', args, filter=filter)


def handle_unpause(args):
    if vars(args).pop('no_history', None):
        filter = get_no_history_filter
    else:
        filter = None
    call('unpause', args, filter=filter)


def call(command, args, payload=None, action='print', filter=None):
    url = args.url + command
    if payload:
        if args.auth_key:
            payload['auth_key'] = args.auth_key
    else:
        payload = {key: (getattr(args, key) if key == 'email'
                         else str(getattr(args, key)))
                   for key in dir(args)
                   if key not in ('host', 'port', 'func', 'url') and
                   not key.startswith('_') and
                   getattr(args, key) is not None and
                   not (key == 'email' and getattr(args, key) == [])}
    response = requests.get(url, params=payload)
    if response.status_code != 200:
        sys.stderr.write('{} {}\n'.format(
            response.status_code, response.reason))
        try:
            sys.exit(pprint.pformat(response.json()).strip())
        except JSONError:
            sys.exit(response.text)
    if action == 'print':
        try:
            content = response.json()
            if filter:
                content = filter(content)
            pprint.pprint(content)
        except BrokenPipeError:  # pragma: no cover
            pass
    elif action == 'return':
        return response.json()
    else:  # pragma: no cover
        raise Exception('Unrecognized action: {}'.format(action))


def periodicity(str):
    if re.match(r'[0-9.]+$', str):
        return float(str)
    return str


def main():  # pragma: no cover
    doit(sys.argv[1:], os.path.expanduser(config_file))


if __name__ == '__main__':  # pragma: no cover
    main()



# Copyright 2015 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.

from coal_mine import cli
from configparser import SafeConfigParser
from io import StringIO
import responses
import subprocess
import sys
import tempfile
from unittest import TestCase
from unittest.mock import patch


def add_response(query, method=responses.GET, host='localhost', port=80,
                 body='{"status": "ok"}', status=200):
    port = '' if port == 80 else ':{}'.format(port)
    url = 'http://{}{}/coal-mine/v1/canary/{}'.format(host, port, query)
    responses.add(method, url, body=body, match_querystring=True,
                  status=status)


class CLITests(TestCase):
    get_body = '{"status": "ok", "canary": {"history": []}}'
    list_body = '{"status": "ok", "canaries": [{"history": []}]}'
    pause_body = '{"status": "ok", "canary": {"paused": true, "history": []}}'
    unpause_body = \
        '{"status": "ok", "canary": {"paused": false, "history": []}}'
    update_body = '{"status": "ok", "canary": {"history": []}}'

    def test_parse_args(self):
        hostname = 'hellohost'
        with tempfile.NamedTemporaryFile(delete=False) as tf:
            args = cli.parse_args(('configure', '--host', hostname),
                                  tf.name)
            self.assertEqual(args.host, hostname)
            self.assertEqual(args.url, 'http://{}/coal-mine/v1/canary/'.format(
                hostname))
            hostname = 'https://' + hostname
            args = cli.parse_args(('configure', '--host', hostname),
                                  tf.name)
            self.assertEqual(args.host, hostname)
            self.assertEqual(args.url, '{}/coal-mine/v1/canary/'.format(
                hostname))

    @responses.activate
    def test_configure(self):
        host = 'hellohost'
        port = '12345'
        auth_key = 'arglebargle'
        with tempfile.NamedTemporaryFile(delete=False) as tf:
            cli.doit(('configure', '--host', host, '--auth-key', auth_key),
                     tf.name)
            config = SafeConfigParser()
            config.read([tf.name])
            self.assertEqual(host, config['coal-mine']['host'])
            with self.assertRaises(KeyError):
                config['coal-mine']['port']
            self.assertEqual(auth_key, config['coal-mine']['auth-key'])

            cli.doit(('configure', '--host', host, '--port', port,
                      '--auth-key', auth_key), tf.name)
            config = SafeConfigParser()
            config.read([tf.name])
            self.assertEqual(host, config['coal-mine']['host'])
            self.assertEqual(port, config['coal-mine']['port'])
            self.assertEqual(auth_key, config['coal-mine']['auth-key'])

            add_response('pause?name=froodle&auth_key=arglebargle',
                         host=host, port=port)
            cli.doit(('pause', '--name', 'froodle'), tf.name)

            cli.doit(('configure', '--no-auth-key'), tf.name)
            config = SafeConfigParser()
            config.read([tf.name])
            with self.assertRaises(KeyError):
                self.assertEqual(auth_key, config['coal-mine']['auth-key'])

            add_response('pause?name=froodle', host=host, port=port)
            cli.doit(('pause', '--name', 'froodle'), tf.name)

            cli.doit(('configure', '--host', host + '2'), tf.name)
            config = SafeConfigParser()
            config.read([tf.name])
            self.assertEqual(host + '2', config['coal-mine']['host'])

    def test_no_command(self):
        with patch('sys.stderr', StringIO()):
            with self.assertRaises(SystemExit):
                cli.doit((), '/dev/null')
            self.assertRegexpMatches(sys.stderr.getvalue(),
                                     r'No command specified')

    @responses.activate
    def test_create(self):
        with patch('sys.stderr', StringIO()):
            with self.assertRaises(SystemExit):
                cli.doit(('create', '--name', 'froodle'), '/dev/null')
            self.assertRegexpMatches(
                sys.stderr.getvalue(),
                r'error: the following arguments are required: --periodicity')

        add_response('create?paused=False&name=froodle&periodicity=60.0')
        cli.doit(('create', '--name', 'froodle', '--periodicity', '60'),
                 '/dev/null')

    @responses.activate
    def test_delete(self):
        add_response('delete?slug=froodle')
        cli.doit(('delete', '--slug', 'froodle'), '/dev/null')

    @responses.activate
    def test_update(self):
        with patch('sys.stderr', StringIO()):
            with self.assertRaises(SystemExit):
                cli.doit(('update', '--id', 'abcdefgh', '--slug', 'foo'),
                         '/dev/null')
            self.assertRegexpMatches(
                sys.stderr.getvalue(),
                r'error: argument --slug: not allowed with argument --id')

        with self.assertRaisesRegexp(
                SystemExit,
                r'Must specify --name, --id, or --slug'):
            cli.doit(('update', '--periodicity', '120'), '/dev/null')

        add_response('update?id=abcdefgh&name=freedle', body=self.update_body)
        cli.doit(('update', '--id', 'abcdefgh', '--name', 'freedle'),
                 '/dev/null')

        add_response('get?name=froodle&auth_key=arglebargle',
                     body='{"status": "ok", "canary": {"id": "bcdefghi"}}')
        add_response('update?description=foodesc&id=bcdefghi&'
                     'auth_key=arglebargle', body=self.update_body)
        cli.doit(('update', '--name', 'froodle', '--description', 'foodesc',
                  '--auth-key', 'arglebargle'),
                 '/dev/null')

        # Same as above but without --auth-key, for code branch coverage,
        # and with history test added.
        add_response('get?name=froodle',
                     body='{"status": "ok", "canary": {"id": "bcdefghi"}}')
        add_response('update?description=foodesc&id=bcdefghi',
                     body=self.update_body)
        with patch('sys.stdout', StringIO()):
            cli.doit(('update', '--name', 'froodle', '--description',
                      'foodesc'), '/dev/null')
            self.assertIn("{'history': []}", sys.stdout.getvalue())

    @responses.activate
    def test_update_no_history(self):
        add_response('get?name=froodle',
                     body='{"status": "ok", "canary": {"id": "bcdefghi"}}')
        add_response('update?description=foodesc&id=bcdefghi',
                     body=self.update_body)
        with patch('sys.stdout', StringIO()):
            cli.doit(('update', '--name', 'froodle', '--description',
                      'foodesc', '--no-history'), '/dev/null')
            self.assertNotIn("{'history': []}", sys.stdout.getvalue())

    @responses.activate
    def test_get(self):
        add_response('get?name=froodle', body=self.get_body)
        with patch('sys.stdout', StringIO()):
            cli.doit(('get', '--name', 'froodle'), '/dev/null')
            self.assertIn("{'history': []}", sys.stdout.getvalue())

    @responses.activate
    def test_get_no_history(self):
        add_response('get?name=froodle', body=self.get_body)
        with patch('sys.stdout', StringIO()):
            cli.doit(('get', '--no-history', '--name', 'froodle'), '/dev/null')
            self.assertNotIn("{'history': []}", sys.stdout.getvalue())

    @responses.activate
    def test_list(self):
        add_response('list')
        cli.doit(('list',), '/dev/null')

        add_response('list?search=foo&late=True&paused=True&verbose=True',
                     body=self.list_body)
        with patch('sys.stdout', StringIO()):
            cli.doit(('list', '--verbose', '--paused', '--late', '--search',
                      'foo'), '/dev/null')
            self.assertIn("{'history': []}", sys.stdout.getvalue())

    @responses.activate
    def test_list_no_history(self):
        add_response('list')
        cli.doit(('list',), '/dev/null')

        add_response('list?search=foo&late=True&paused=True&verbose=True',
                     body=self.list_body)
        with patch('sys.stdout', StringIO()):
            cli.doit(('list', '--verbose', '--no-history', '--paused',
                      '--late', '--search', 'foo'), '/dev/null')
            self.assertNotIn("{'history': []}", sys.stdout.getvalue())

    @responses.activate
    def test_trigger(self):
        add_response('trigger?name=froodle')
        cli.doit(('trigger', '--name', 'froodle'), '/dev/null')

    @responses.activate
    def test_pause(self):
        add_response('pause?name=froodle', body=self.pause_body)
        with patch('sys.stdout', StringIO()):
            cli.doit(('pause', '--name', 'froodle'), '/dev/null')
            self.assertIn("'history': []", sys.stdout.getvalue())

    @responses.activate
    def test_pause_no_history(self):
        add_response('pause?name=froodle', body=self.pause_body)
        with patch('sys.stdout', StringIO()):
            cli.doit(('pause', '--no-history', '--name', 'froodle'),
                     '/dev/null')
            self.assertNotIn("'history': []", sys.stdout.getvalue())

    @responses.activate
    def test_unpause(self):
        add_response('unpause?name=froodle', body=self.unpause_body)
        with patch('sys.stdout', StringIO()):
            cli.doit(('unpause', '--name', 'froodle'), '/dev/null')
            self.assertIn("'history': []", sys.stdout.getvalue())

    @responses.activate
    def test_unpause_no_history(self):
        add_response('unpause?name=froodle', body=self.unpause_body)
        with patch('sys.stdout', StringIO()):
            cli.doit(('unpause', '--no-history', '--name', 'froodle'),
                     '/dev/null')
            self.assertNotIn("'history': []", sys.stdout.getvalue())

    @responses.activate
    def test_bad_response(self):
        with self.assertRaisesRegexp(SystemExit, r'^Not Found$'):
            add_response('get?name=froodle', status=404, body='Not Found')
            cli.doit(('get', '--name', 'froodle'), '/dev/null')

    def test_periodicity_string(self):
        val = '* * * * * 65'
        self.assertEqual(val, cli.periodicity(val))

    def test_subprocess_invocation(self):
        with self.assertRaises(subprocess.CalledProcessError) as cm:
            subprocess.check_output(
                ('python', '-m', 'coal_mine.cli', 'froodle'),
                stderr=subprocess.STDOUT)
        self.assertRegexpMatches(cm.exception.output.decode('ascii'),
                                 r"invalid choice: 'froodle'")



# Copyright 2016 Quantopian, Inc.
# Copyright 2022 Jonathan Kamens
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.

from coal_mine.mongo_store import MongoStore
from pymongo import MongoClient, IndexModel, ASCENDING
from pymongo.errors import OperationFailure
from unittest import TestCase
import uuid


class MongoStoreInitTests(TestCase):
    def setUp(self):
        self.db_hosts = ['localhost']
        self.db_name = "coal-mine-test-" + str(uuid.uuid4())
        self.db_conn = MongoClient()
        self.db = self.db_conn[self.db_name]

    def tearDown(self):
        self.db_conn.drop_database(self.db)

    def test_basic_init(self):
        MongoStore(self.db_hosts, self.db_name, None, None)
        pass

    def test_init_auth(self):
        with self.assertRaises(OperationFailure):
            MongoStore(self.db_hosts, self.db_name, 'nope', 'nope')

    def test_init_index_compatibility(self):
        collection = self.db['canaries']
        collection.create_indexes([IndexModel([('id', ASCENDING)])])
        existing_indexes = collection.index_information()
        self.assertNotIn('unique', existing_indexes['id_1'])
        MongoStore(self.db_hosts, self.db_name, None, None)
        new_existing_indexes = collection.index_information()
        self.assertTrue(new_existing_indexes['id_1']['unique'])
        # For code coverage, testing when is_unique is already True as it
        # should be.
        MongoStore(self.db_hosts, self.db_name, None, None)


class MongoStoreTests(TestCase):
    def setUp(self):
        self.db_hosts = ['localhost']
        self.db_name = "coal-mine-test-" + str(uuid.uuid4())
        self.db_conn = MongoClient()
        self.db = self.db_conn[self.db_name]
        self.store = MongoStore(self.db_hosts, self.db_name, None, None)

    def tearDown(self):
        self.db_conn.drop_database(self.db)

    def test_create(self):
        self.store.create({'id': 'abcdefgh'})

    def test_update_noop(self):
        self.store.create({'id': 'abcdefgh'})
        self.store.update('abcdefgh', {})

    def test_update_set(self):
        self.store.create({'id': 'abcdefgh'})
        self.store.update('abcdefgh', {'periodicity': 10})
        self.assertEqual(self.store.get('abcdefgh')['periodicity'], 10)

    def test_update_unset(self):
        self.store.create({'id': 'abcdefgh'})
        self.store.update('abcdefgh', {'periodicity': 10})
        self.store.update('abcdefgh', {'periodicity': None})
        self.assertNotIn('periodicity', self.store.get('abcdefgh'))

    def test_get_nonexistent(self):
        with self.assertRaises(KeyError):
            self.store.get('abcdefgh')

    def test_list(self):
        self.store.create({'id': 'abcdefgh', 'name': 'freedle',
                           'periodicity': 600, 'paused': False,
                           'late': False})
        iterator = self.store.list()
        next(iterator)
        with self.assertRaises(StopIteration):
            next(iterator)
        next(self.store.list(verbose=True))
        next(self.store.list(paused=False))
        next(self.store.list(late=False))
        next(self.store.list(order_by='deadline'))
        next(self.store.list(search=r'freedle'))

    def test_upcoming_deadlines(self):
        self.store.create({'id': 'abcdefgh', 'paused': False, 'late': False})
        next(self.store.upcoming_deadlines())

    def test_delete(self):
        self.store.create({'id': 'abcdefgh'})
        self.store.delete('abcdefgh')
        with self.assertRaises(KeyError):
            self.store.delete('abcdefgh')

    def test_find_identifier(self):
        self.store.create({'id': 'abcdefgh', 'slug': 'froodle'})
        self.assertEqual(self.store.find_identifier('froodle'), 'abcdefgh')
        with self.assertRaises(KeyError):
            self.store.find_identifier('freedle')



# Copyright 2015 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.

from coal_mine.crontab_schedule import \
    CronTabSchedule, CronTabScheduleException, FastCronTab
from datetime import datetime, timedelta
import time
from unittest import TestCase

SCHEDULE = '''
# This is a comment.
    *          * * * Sun 300

# Leave the blank line above and this comment in place.
    * 22-23,0-12 * * Mon-Fri 300
 0-29         13 * * Mon-Fri 300
30-59         13 * * Mon-Fri 90
    *      14-21 * * Mon-Fri 90
'''

MULTI_SCHEDULE = '''
* 0-11 * * * A
* 7-18 * * * B
* 12-23 * * * C
'''

GAP_SCHEDULE = '''
* 12-23 * * * A
'''

GAP_EXPECTED = [
    (datetime(2015, 1, 1, 0, 0), datetime(2015, 1, 1, 11, 59), None),
    (datetime(2015, 1, 1, 12, 0), datetime(2015, 1, 1, 12, 1), ('A',))]

FIXED_EXPECTED = [
    (datetime(2015, 1, 1, 0, 0), datetime(2015, 1, 1, 13, 29), '300'),
    (datetime(2015, 1, 1, 13, 30), datetime(2015, 1, 1, 21, 59), '90'),
    (datetime(2015, 1, 1, 22, 0), datetime(2015, 1, 2, 13, 29), '300'),
    (datetime(2015, 1, 2, 13, 30), datetime(2015, 1, 2, 21, 59), '90'),
    (datetime(2015, 1, 2, 22, 0), datetime(2015, 1, 2, 23, 59), '300'),
    (datetime(2015, 1, 3, 0, 0), datetime(2015, 1, 3, 23, 59), None),
    (datetime(2015, 1, 4, 0, 0), datetime(2015, 1, 5, 13, 29), '300'),
    (datetime(2015, 1, 5, 13, 30), datetime(2015, 1, 5, 21, 59), '90'),
    (datetime(2015, 1, 5, 22, 0), datetime(2015, 1, 6, 13, 29), '300'),
    (datetime(2015, 1, 6, 13, 30), datetime(2015, 1, 6, 21, 59), '90'),
    (datetime(2015, 1, 6, 22, 0), datetime(2015, 1, 7, 13, 29), '300'),
    (datetime(2015, 1, 7, 13, 30), datetime(2015, 1, 7, 21, 59), '90'),
    (datetime(2015, 1, 7, 22, 0), datetime(2015, 1, 8, 0, 0), '300')
]

DYNAMIC_EXPECTED = [
    (datetime(2015, 1, 1, 0, 0), datetime(2015, 1, 1, 13, 29), '300'),
    (datetime(2015, 1, 1, 13, 30), datetime(2015, 1, 1, 21, 59), '90'),
    (datetime(2015, 1, 1, 22, 0), datetime(2015, 1, 2, 13, 29), '300'),
    (datetime(2015, 1, 2, 13, 30), datetime(2015, 1, 2, 21, 59), '90'),
    (datetime(2015, 1, 2, 22, 0), datetime(2015, 1, 2, 23, 59), '300'),
    (datetime(2015, 1, 3, 0, 0), datetime(2015, 1, 3, 23, 59), None)
]

MULTI_EXPECTED = [
    (datetime(2015, 1, 1, 0, 0), datetime(2015, 1, 1, 6, 59), ('A',)),
    (datetime(2015, 1, 1, 7, 0), datetime(2015, 1, 1, 11, 59), ('A', 'B')),
    (datetime(2015, 1, 1, 12, 0), datetime(2015, 1, 1, 18, 59), ('B', 'C'))
]


class CronTabScheduleTests(TestCase):
    maxDiff = None

    def test_init(self):
        s = CronTabSchedule(SCHEDULE)
        self.assertEqual(len(s.entries), 5)
        self.assertEqual(s.entries[2][1], '300')
        # The rest is thrown in here for code coverage.
        s.next_minute()
        s.next_minute(now=datetime.now())

    def test_schedule_iter_fixed(self):
        s = CronTabSchedule(SCHEDULE)
        slots = [slot for slot in s.schedule_iter(
            start=datetime(2015, 1, 1),
            end=datetime(2015, 1, 8),
            multi=False)]
        self.assertEqual(slots, FIXED_EXPECTED)
        # Code coverage.
        next(s.schedule_iter())

    def test_schedule_iter_dynamic(self):
        s = CronTabSchedule(SCHEDULE)
        slots = [slot for slot in s.schedule_iter(
            start=datetime(2015, 1, 1),
            multi=False)]
        self.assertEqual(slots, DYNAMIC_EXPECTED)

    def test_schedule_iter_gap(self):
        s = CronTabSchedule(GAP_SCHEDULE)
        slots = [slot for slot in s.schedule_iter(
            start=datetime(2015, 1, 1),
            end=datetime(2015, 1, 1, 12, 1))]
        self.assertEqual(slots, GAP_EXPECTED)

    def test_schedule_iter_end_endless(self):
        s = CronTabSchedule(SCHEDULE)
        with self.assertRaises(ValueError):
            next(s.schedule_iter(end=datetime.now(), endless=True))

    def test_multi_ok(self):
        s = CronTabSchedule(MULTI_SCHEDULE)
        slots = [slot for slot in s.schedule_iter(
            start=datetime(2015, 1, 1))]
        self.assertEqual(slots, MULTI_EXPECTED)

    def test_multi_exception(self):
        s = CronTabSchedule(MULTI_SCHEDULE)
        with self.assertRaises(CronTabScheduleException):
            [slot for slot in s.schedule_iter(
                start=datetime(2015, 1, 1),
                multi=False)]

    def test_add_entry_invalid(self):
        with self.assertRaises(CronTabScheduleException):
            CronTabSchedule('* * *')

    def test_delimiter(self):
        # Paranoia: Make sure we're actually testing what we think
        # we're testing, by ensuring that the original schedule had
        # newlines in it and the one we're using doesn't.
        self.assertNotEqual(SCHEDULE.find('\n'), -1)
        with_semis = SCHEDULE.replace('\n', ';')
        self.assertEqual(with_semis.find('\n'), -1)
        s = CronTabSchedule(with_semis, delimiter=';')
        slots = [slot for slot in s.schedule_iter(
            start=datetime(2015, 1, 1),
            multi=False)]
        self.assertEqual(slots, DYNAMIC_EXPECTED)

    def test_no_schedule(self):
        s = CronTabSchedule('')
        with self.assertRaises(CronTabScheduleException):
            s.soonest()

    def test_soonest_no_now(self):
        s = CronTabSchedule('* * * * * foo\n* * * * * bar')
        now = datetime.now()
        if not now.second:
            time.sleep(1)
            now = datetime.now()
        nxt = s.soonest()
        self.assertEqual(nxt, now + timedelta(
            seconds=60 - now.second - now.microsecond / 1000000))

    def test_soonest_with_seconds(self):
        s = CronTabSchedule('* * * * * foo')
        now = datetime(2015, 1, 1, 1, 1, 1, 1)
        nxt = s.soonest(now=now)
        self.assertEqual(nxt, datetime(2015, 1, 1, 1, 2))

    def test_soonest_backtrack_needed(self):
        s = CronTabSchedule('* * * * * foo')
        now = datetime(2015, 1, 1)
        nxt = s.soonest(now=now)
        self.assertEqual(now, nxt)

    def test_FastCronTab_default_now(self):
        e = FastCronTab('* * * * *')
        now = datetime.now()
        # Avoid race conditions. What happens if between when we call
        # datetime.now() and when crontab_schedule.py calls it, the
        # minute rolls over?
        if now.second > 55:
            time.sleep(60 - now.second + 1)
            now = datetime.now()
        self.assertAlmostEqual(e.next(), e.next(now), places=2)

    def test_FastCronTab_default_utc(self):
        e = FastCronTab('* * * * *')
        self.assertNotEqual(e.next(default_utc=False),
                            e.next(default_utc=True))

    def test_FastCronTab_default_utc_unspecified(self):
        e = FastCronTab('*/5 * * * *')
        now = datetime.now()
        first = e.next(now)
        second = e.next(now + timedelta(seconds=first + 1), default_utc=False)
        self.assertEqual(second, 299.0)
